{-# LANGUAGE ScopedTypeVariables #-}

module Main where

import Control.Monad (forM)
import qualified Data.ByteString.Lazy as L
import Data.Git.Ref (Ref, SHA1, hashLBS)
import Data.Git.Repository (Git)
import Data.Git.Storage
  ( Git (gitRepoPath),
    findRepoMaybe,
    withCurrentRepo,
  )
import Data.Git.Storage.Loose
  ( looseEnumeratePrefixes,
    looseEnumerateWithPrefix,
    looseMarshall,
    looseRead,
    looseReadRaw,
    looseUnmarshallRaw,
  )
import Data.Git.Storage.Object (ObjectHeader)
import Data.Maybe (catMaybes)
import Monad (testGitMonadLocal)
import System.Exit (exitFailure)
import Text.Bytedump (dumpDiffLBS)

onLocalRepo :: (Git SHA1 -> IO ()) -> IO ()
onLocalRepo f = do
  fpath <- findRepoMaybe
  case fpath of
    Nothing -> putStrLn "cannot run this test without repository. clone the original repository for test"
    Just _ -> withCurrentRepo f

doLocalMarshallEq ::
  Git SHA1 ->
  IO [[Maybe (Ref SHA1, Ref SHA1, (ObjectHeader SHA1, L.ByteString), (ObjectHeader SHA1, L.ByteString))]]
doLocalMarshallEq git = do
  prefixes <- looseEnumeratePrefixes (gitRepoPath git)
  forM prefixes $ \prefix -> do
    refs <- looseEnumerateWithPrefix (gitRepoPath git) prefix
    forM refs $ \ref -> do
      raw <- looseReadRaw (gitRepoPath git) ref
      obj <- looseRead (gitRepoPath git) ref
      let content = looseMarshall obj
      let raw2 = looseUnmarshallRaw content
      let hashed = hashLBS content
      if ref /= hashed
        then return $ Just (ref, hashed, raw, raw2)
        else return Nothing

printDiff :: (Show a1, Show a2, Show a3, Show a4) => (a1, a2, (a3, L.ByteString), (a4, L.ByteString)) -> IO ()
printDiff (actualRef, gotRef, (actualHeader, actualRaw), (gotHeader, gotRaw)) = do
  putStrLn "=========== difference found"
  putStrLn ("ref expected: " ++ show actualRef)
  putStrLn ("ref got     : " ++ show gotRef)
  putStrLn ("header expected: " ++ show actualHeader)
  putStrLn ("header got     : " ++ show gotHeader)
  putStrLn "raw diff:"
  putStrLn $ dumpDiffLBS actualRaw gotRaw

printLocalMarshallError :: (Foldable t, Show a1, Show a2, Show a3, Show a4) => t (a1, a2, (a3, L.ByteString), (a4, L.ByteString)) -> IO ()
printLocalMarshallError l
  | null l = putStrLn "local marshall:   [OK]"
  | otherwise =
    putStrLn ("local marshall: [" ++ show (length l) ++ " errors]")
      >> mapM_ printDiff l
      >> exitFailure

main :: IO ()
main = do
  onLocalRepo $ \(git :: Git SHA1) -> do
    doLocalMarshallEq git >>= printLocalMarshallError . catMaybes . concat
    testGitMonadLocal
    return ()
