{-# LANGUAGE FlexibleInstances #-}
{-# OPTIONS_GHC -fno-warn-orphans #-}

module Main where

import Control.Monad (liftM3, replicateM)
import qualified Data.ByteString as B
import qualified Data.ByteString.Lazy as L
import Data.Git.Ref
  ( HashAlgorithm (..),
    Ref,
    SHA1,
    fromBinary,
    fromHex,
    toBinary,
    toHex,
  )
import Data.Git.Revision
  ( RevModifier
      ( RevModAtDate,
        RevModAtType,
        RevModParent,
        RevModParentFirstN
      ),
    Revision (..),
    fromString,
  )
import Data.Git.Storage.Loose (looseMarshall, looseUnmarshall)
import Data.Git.Storage.Object
  ( Object,
    ObjectType (TypeBlob, TypeCommit, TypeTag, TypeTree),
    Objectable (toObject),
  )
import Data.Git.Types
  ( Blob (Blob),
    Commit (Commit),
    CommitExtra (CommitExtra),
    EntName,
    GitTime (GitTime),
    ModePerm (..),
    Person (Person),
    Tag (Tag),
    Tree (Tree),
    TreeEnt,
    entName,
  )
import Data.Hourglass
  ( Elapsed (..),
    Seconds (Seconds),
    TimezoneOffset (TimezoneOffset),
  )
import Test.Tasty (TestTree, defaultMain, testGroup)
import Test.Tasty.QuickCheck
  ( Arbitrary (arbitrary),
    Gen,
    Positive (getPositive),
    choose,
    elements,
    frequency,
    oneof,
    resize,
    testProperty,
  )

-- for arbitrary instance to generate only data that are writable
-- to disk. i.e. no deltas.
data ObjNoDelta hash = ObjNoDelta (Object hash)
  deriving (Eq)

instance Show (ObjNoDelta hash) where
  show (ObjNoDelta o) = show o

arbitraryBS :: Int -> Gen B.ByteString
arbitraryBS size = B.pack . map fromIntegral <$> replicateM size (choose (0, 255) :: Gen Int)

arbitraryBSno0 :: Int -> Gen B.ByteString
arbitraryBSno0 size = B.pack . map fromIntegral <$> replicateM size (choose (1, 255) :: Gen Int)

arbitraryBSasciiNoSpace :: Int -> Gen B.ByteString
arbitraryBSasciiNoSpace size = B.pack . map fromIntegral <$> replicateM size (choose (0x21, 0x7f) :: Gen Int)

arbitraryBSascii :: Int -> Gen B.ByteString
arbitraryBSascii size = B.pack . map fromIntegral <$> replicateM size (choose (0x20, 0x7f) :: Gen Int)

arbitraryBSnoangle :: Int -> Gen B.ByteString
arbitraryBSnoangle size = B.pack . map fromIntegral <$> replicateM size (choose (0x40, 0x7f) :: Gen Int)

arbitraryEntname :: Int -> Gen EntName
arbitraryEntname size = entName . B.pack . map fromIntegral <$> replicateM size range
  where
    range :: Gen Int
    range =
      oneof
        [ choose (0x21, 0x2e), -- remove 0x2f (slash)
          choose (0x30, 0x7f)
        ]

arbitraryRef :: HashAlgorithm hash => hash -> Gen (Ref hash)
arbitraryRef alg = fromBinary <$> arbitraryBS (hashDigestSize alg)

instance HashAlgorithm hash => Arbitrary (Ref hash) where
  arbitrary = arbitraryRef (error "alg")

arbitraryMsg :: Gen B.ByteString
arbitraryMsg = arbitraryBSno0 1

arbitraryLazy :: Gen L.ByteString
arbitraryLazy = L.fromChunks . (: []) <$> arbitraryBS 40

arbitraryRefList :: Gen [Ref SHA1]
arbitraryRefList = replicateM 2 arbitrary

arbitraryEnt :: Gen (TreeEnt SHA1)
arbitraryEnt = liftM3 (,,) arbitrary (arbitraryEntname 23) arbitrary

arbitraryEnts :: Gen [TreeEnt SHA1]
arbitraryEnts = choose (1, 2) >>= \i -> replicateM i arbitraryEnt

instance Arbitrary TimezoneOffset where
  arbitrary = TimezoneOffset <$> choose (-11 * 60, 12 * 60)

instance Arbitrary Elapsed where
  arbitrary = Elapsed . Seconds <$> choose (0, 2 ^ 32 -1)

instance Arbitrary GitTime where
  arbitrary = GitTime <$> arbitrary <*> arbitrary

instance Arbitrary ModePerm where
  arbitrary = ModePerm <$> elements [0o644, 0o664, 0o755, 0]

instance Arbitrary RevModifier where
  arbitrary =
    oneof
      [ RevModParent . getPositive <$> arbitrary,
        RevModParentFirstN . getPositive <$> arbitrary,
        RevModAtType <$> arbitraryType,
        RevModAtDate <$> arbitraryDate
        --, RevModAtN . getPositive <$> arbitrary
      ]

arbitraryDate :: Gen [Char]
arbitraryDate = elements ["yesterday", "29-Jan-1982", "5 days ago"]

arbitraryType :: Gen [Char]
arbitraryType = elements ["commit", "tree"]

instance Arbitrary Revision where
  arbitrary = do
    s <- choose (1, 40) >>= flip replicateM (elements ['a' .. 'z'])
    rms <- choose (1, 4) >>= flip replicateM arbitrary
    return $ Revision s rms

arbitraryName :: Gen Person
arbitraryName =
  liftM3
    Person
    (arbitraryBSnoangle 16)
    (arbitraryBSnoangle 16)
    arbitrary

arbitraryObjTypeNoDelta :: Gen ObjectType
arbitraryObjTypeNoDelta = oneof [return TypeTree, return TypeBlob, return TypeCommit, return TypeTag]

arbitrarySmallList :: Gen [CommitExtra]
arbitrarySmallList = frequency [(2, return []), (1, resize 3 arbitrary)]

instance Arbitrary (Commit SHA1) where
  arbitrary = Commit <$> arbitrary <*> arbitraryRefList <*> arbitraryName <*> arbitraryName <*> return Nothing <*> arbitrarySmallList <*> arbitraryMsg

instance Arbitrary CommitExtra where
  arbitrary = CommitExtra <$> arbitraryBSasciiNoSpace 80 <*> arbitraryMsg

instance Arbitrary (Tree SHA1) where
  arbitrary = Tree <$> arbitraryEnts

instance Arbitrary (Blob SHA1) where
  arbitrary = Blob <$> arbitraryLazy

instance Arbitrary (Tag SHA1) where
  arbitrary = Tag <$> arbitrary <*> arbitraryObjTypeNoDelta <*> arbitraryBSascii 20 <*> arbitraryName <*> arbitraryMsg

instance Arbitrary (ObjNoDelta SHA1) where
  arbitrary =
    ObjNoDelta
      <$> oneof
        [ toObject <$> (arbitrary :: Gen (Commit SHA1)),
          toObject <$> (arbitrary :: Gen (Tree SHA1)),
          toObject <$> (arbitrary :: Gen (Blob SHA1)),
          toObject <$> (arbitrary :: Gen (Tag SHA1))
        ]

prop_object_marshalling_id :: ObjNoDelta SHA1 -> Bool
prop_object_marshalling_id (ObjNoDelta obj) =
  let unmarshall = looseUnmarshall :: L.ByteString -> Object SHA1
   in obj `assertEq` (unmarshall $ looseMarshall obj)
  where
    assertEq a b
      | show a == show b = True
      | otherwise = error ("not equal:\n" ++ show a ++ "\ngot: " ++ show b)

refTests :: [TestTree]
refTests =
  [ testProperty "hexadecimal" (marshEqual (fromHex . toHex :: Ref SHA1 -> Ref SHA1)),
    testProperty "binary" (marshEqual (fromBinary . toBinary :: Ref SHA1 -> Ref SHA1)),
    testProperty "ref" $ marshEqual (fromString . show :: Revision -> Revision)
  ]
  where
    marshEqual t ref = ref `assertEq` t ref
    assertEq a b
      | a == b = True
      | otherwise = error ("expecting: " ++ show a ++ " got: " ++ show b)

objTests :: [TestTree]
objTests =
  [ testProperty "unmarshall.marshall==id" prop_object_marshalling_id
  ]

main :: IO ()
main =
  defaultMain $
    testGroup
      "hgit"
      [ testGroup "ref marshalling" refTests,
        testGroup "object marshalling" objTests
      ]
